/*
*
* Gtk Utility routines: Text View Widget.
* Copyright (C) 2002-2006 by Tito Ragusa <tito-wolit@tiscali.it>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License as
* published by the Free Software Foundation; either version 2 of
* the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*
* Please read the COPYING and README file!!!
*
* Report bugs to <tito-wolit@tiscali.it>
*
*/

#include <gtk/gtk.h>
#include <sys/types.h>
#include <sys/stat.h>

/********************* Text file View Widget *******************************/

static int view_lock = 0;
static GtkWidget *view_window;
static GtkWidget *view_text;

static void text_view_quit_callback(void)
{
	if (view_lock) return;

	gtk_main_quit();
	gtk_widget_destroy(view_window);
}

static void text_view_wrap(void)
{
	if (view_lock) return;

	gtk_text_view_set_wrap_mode (GTK_TEXT_VIEW(view_text),
				(gtk_text_view_get_wrap_mode (GTK_TEXT_VIEW(view_text)) != GTK_WRAP_NONE) ?
				GTK_WRAP_NONE :
				GTK_WRAP_WORD);
}


static GtkActionEntry text_view_entries[] = {
	/* name,		stock id,	label,				accelerator,		tooltip,	callback */
	{ "FileMenu",	NULL,		MNU_STR_FILE,		NULL,				NULL,		NULL},
	{ "OptionsMenu",NULL,		MNU_STR_OPTIONS,	NULL,				NULL,		NULL},	
	{ "HelpMenu",	NULL,		MNU_STR_HELP,		NULL,				NULL,		NULL},
	{ "Quit",		NULL,		MNU_STR_QUIT,		SHC_FILE_QUIT,		NULL,		G_CALLBACK(text_view_quit_callback) },
	{ "About",		NULL,		MNU_STR_ABOUT,		SHC_HELP_ABOUT,		NULL,		G_CALLBACK(about_cb) },
};

static GtkToggleActionEntry text_view_check_entries[] = {
{	
	"Wrap", NULL,								/* name, stock id */
	MNU_STR_WRAP, SHC_OPTIONS_WRAP,				/* label, accelerator */     
	NULL,										/* tooltip */
	G_CALLBACK (text_view_wrap),				/* callback */ 
	FALSE },									/* is_active */
};

static const gchar *text_view_ui_info = 
"<ui>"
"  <menubar name='MenuBar'>"
"    <menu action='FileMenu'>"
"      <menuitem action='Quit'/>"
"    </menu>"
"    <menu action='OptionsMenu'>"
"      <menuitem action='Wrap'/>"
"    </menu>"
"    <menu action='HelpMenu'>"
"      <menuitem action='About'/>"
"    </menu>"
"  </menubar>"
"</ui>";

static struct MenuEntry text_view_menu_entries[] = {
{ text_view_entries,	NULL,						NULL,	G_N_ELEMENTS (text_view_entries),			0, NULL},
{ NULL,					text_view_check_entries,	NULL,	G_N_ELEMENTS (text_view_check_entries),		0, NULL},
{ NULL,					NULL,						NULL,	0,											0, NULL},
/* Set last to NULL */ 
};

void textview_window_create(char * title, int x, int y, int border, char *filename, FILE *fd)
{
	GtkWidget *menubar;
	GtkWidget *table;
	ProgressData *pdata;
	FILE *file;
	struct stat st;

	DBG_FUNCTION_NAME;
	
	view_window = window_create(x, y, border, title, GTK_SIGNAL_FUNC(text_view_quit_callback), NULL, MODAL);
	table =  table_create(20, 21,GTK_WIDGET(view_window), 0, HOMOGENEUS);
	menubar = main_menu_create(GTK_WIDGET (view_window), text_view_menu_entries, text_view_ui_info, NULL);
	gtk_table_attach_defaults (GTK_TABLE(table),GTK_WIDGET(menubar), 0, 20, 0, 1);
	view_text = textpad_create_on_table(NOT_EDITABLE, NULL, NULL, NO_WRAP, SCROLL, table, 0, 20, 1, 20);
	pdata = progressbar_create_on_table((fd) ? ACTIVITY : PULSE, table, 0, 10, 20, 21);
	gtk_widget_show_all(GTK_WIDGET(view_window));
	view_lock = 1;
	if (fd) {
		read_from_stream_and_display_text(view_text, fd, 0, NULL);
	} else if ((file = open_file(filename)) != NULL &&	stat(filename, &st) == 0 ) {
		read_from_stream_and_display_text(view_text, file, st.st_size, &pdata->value);
		xfclose_nostdin(file);
	}
	progress_destroy(pdata);
	view_lock = 0;
	gtk_main();
}

/***************End of Text file View Widget ********************************/
