/*
*
* Gtk+2.x  Utility routines: Image Viewer.
* Copyright (C) 2002-2006 by Tito Ragusa <tito-wolit@tiscali.it>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License as
* published by the Free Software Foundation; either version 2 of
* the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*
* Please read the COPYING and README file!!!
*
* Report bugs to <tito-wolit@tiscali.it>
*
*/

/* This is based gview-0.9.1a (http://gview.sourceforge.net/)
   by Jason Barto <jpbarto_at_sourceforge.net> 
   released under the  GNU General Public License (GPL),
   and on pho-0.9.2 (http://www.shallowsky.com/software/pho/)
   by Akkana Peck <askakk@shallowsky.com>
   released under the  GNU General Public License (GPL).
 */



#include <gtk/gtk.h>
#include <gdk-pixbuf/gdk-pixbuf.h>

GdkPixbuf * image_resize(GdkPixbuf *img, gint new_width, gint new_height)
{
	gint tmp_gdk_height=0;
	gint tmp_gdk_width=0;
	gint width;
	gint height;
	float w_ratio;
	float h_ratio;
	
	width  = gdk_pixbuf_get_width  (img);
	height = gdk_pixbuf_get_height (img);
		
	/* check that the image is not bigger than allowed */
	/* if the image is too tall then reset */
	/* the tmp_gdk_height to reflect this */
	if (height > new_height)
		tmp_gdk_height = new_height;
	
	if (width > new_width)
		tmp_gdk_width = new_width;
	
	/* if image is too big, resize it */
	/* if either tmp variables were set then the image DOES exceed */
	/* the allowed size and we must resize it to fit */
	if (tmp_gdk_height > 0 || tmp_gdk_width > 0) {
		/* calculate the width to height ratio */
		w_ratio = width / new_width;
		h_ratio = height / new_height;
	
		/* calculate new gdkpixbuf measurements */
		/* determine if the image is wider than tall or vice versa */
		/* calculate a new height or width that keeps the image ratio */
		if (w_ratio >= h_ratio) {
			tmp_gdk_height = ((float)height / (float)width) * tmp_gdk_width;
		} else {
			tmp_gdk_width  = ((float)width / (float)height) * tmp_gdk_height;
		}
	
		img = gdk_pixbuf_scale_simple (img, tmp_gdk_width, tmp_gdk_height, GDK_INTERP_BILINEAR);
	}
	return img;
}

GdkPixbuf * image_fit_to_screen (GdkPixbuf *img)
{

	int screen_height;
	int screen_width;
	GdkScreen *screen;
	
	/* get a handle to the screen for its measurements */
	screen = gdk_screen_get_default ();
	/* using the gdk library functions, get the height and width */
	screen_height = gdk_screen_get_height (screen);
	screen_width = gdk_screen_get_width (screen);
	/* leave some buffer room around our image window */
	/* allowing room for things like taskbars that may be on the screen */
	screen_height -= SCREEN_MARGIN;
	screen_width -= SCREEN_MARGIN;
	
	return image_resize(img, screen_width, screen_height);
}

void ViewImageWindow(char *filename)
{
	GtkWidget *win;
	GdkPixbuf *image;

	/* create the image, the GdkPixbuf structure is used store the image data */
	if (filename == NULL || (image=gdk_pixbuf_new_from_file (filename, NULL	)) == NULL) {
		dialog_window(STR_ERROR, ONE_BUTTON, NULL, 1, STR_CANNOT_OPEN_FILE"\n%s", filename);
		return;
	}
	
	image =  image_fit_to_screen (image);
	
	win = window_create (gdk_pixbuf_get_width (image),
						gdk_pixbuf_get_height(image),
						0, filename,  gtk_main_quit, NULL, MODAL);
	gtk_container_add (GTK_CONTAINER (win), gtk_image_new_from_pixbuf (image));
	main_loop(&win);
}
